from __future__ import absolute_import, print_function, unicode_literals

from unittest import TestCase
try:
    from unittest import mock
except ImportError:
    import mock

from ssr.utils import SecretKeyInterface


class TestSecretKeyInterface(TestCase):

    def test_get_secret_key(self):
        interface = SecretKeyInterface()
        with mock.patch.object(interface, 'get_secret_key') as get_secret_key:
            secret_key = interface.secret_key
            self.assertEqual(secret_key, get_secret_key())
        self.assertIsNone(interface.get_secret_key())

    def test_secret_key_idempotence(self):
        interface = SecretKeyInterface()
        with mock.patch.object(interface, 'get_secret_key') as get_secret_key:
            interface.secret_key
            self.assertEqual(get_secret_key.call_count, 1)
            interface.secret_key
            self.assertEqual(get_secret_key.call_count, 1)

    def test_secret_key_setter(self):
        interface = SecretKeyInterface()
        with mock.patch.object(interface, 'get_secret_key') as get_secret_key:
            interface.secret_key = 'foo'
            self.assertEqual(interface.secret_key, 'foo')
            self.assertEqual(get_secret_key.call_count, 0)



from __future__ import absolute_import, print_function, unicode_literals


class Error(Exception):
    """Simple base error for the package."""

    pass


class AuthenticationError(Error):
    """AuthenticationError indicates either invalid or expired credentials."""

    pass


class TimeoutError(AuthenticationError):
    """Invalid authentication timestamp."""

    pass



from __future__ import absolute_import, print_function, unicode_literals

import time

try:  # python2
    from functools32 import lru_cache
except ImportError:  # pragma: no cover python3
    from functools import lru_cache

from . import utils


class Client(utils.SecretKeyInterface):

    def __init__(self, client_id, secret_key=None, auth_type=None, timeout=300):
        self.id = client_id
        self.secret_key = secret_key
        self.auth_type = auth_type or 'SSR'
        self.timestamp = int(time.time())
        self.timeout = timeout

    @lru_cache(maxsize=1)
    def generate_pub_key(self, timestamp):
        assert self.secret_key is not None
        return utils.generate_public_key(self.id, timestamp, self.secret_key)

    def get_timestamp(self):
        now = int(time.time())
        if not 0 < now - self.timestamp <= self.timeout:
            self.timestamp = now
        return self.timestamp

    @property
    def headers(self):
        """Return a dictionary containing the headers to use in a request.

        Contains:
            client_id {string}
            timestamp {string}
            authorization {string}
        """
        timestamp = self.get_timestamp()
        return self.generate_headers(timestamp)

    @property
    def access_key(self):
        timestamp = self.get_timestamp()
        return '{}:{}:{}'.format(
            self.id,
            timestamp,
            self.generate_pub_key(timestamp)
        )

    def generate_headers(self, timestamp):
        pub_key = self.generate_pub_key(timestamp)
        return {
            'client_id': self.id,
            'timestamp': str(timestamp),
            'authorization': '{} {}'.format(self.auth_type, pub_key)
        }

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self)

    def __str__(self):
        return self.id



from __future__ import absolute_import, print_function, unicode_literals

import time
from .exceptions import AuthenticationError, TimeoutError
from . import utils


class BaseAuthentication(utils.SecretKeyInterface):

    auth_type = 'SSR'
    timeout = 300
    header_prefix = None
    access_key_name = 'key'

    def authenticate_header(self, request):
        # For compatability with django rest framework authentication classes.
        pass

    def get_header_value(self, data, name, default=None):
        prefix = self.header_prefix or ''
        name = prefix + name
        return data.get(name.lower(), default)

    def validate_headers(self, data):
        cleaned_data = {key.lower(): value for key, value in data.items()}
        authorization = self.get_header_value(cleaned_data, 'authorization', '')
        if not authorization.startswith('{} '.format(self.auth_type)):
            return
        parts = authorization.split(' ')
        if len(parts) != 2:
            return False, AuthenticationError('Invalid authorization.')
        _, pub_key = parts

        client_id = self.get_header_value(cleaned_data, 'client_id')
        timestamp = self.get_header_value(cleaned_data, 'timestamp', 0)
        return self.validate(client_id, timestamp, pub_key)

    def validate_access_key(self, access_key):
        try:
            client_id, timestamp, pub_key = access_key.split(':')
        except ValueError:
            return False, AuthenticationError('Invalid access key.')
        return self.validate(client_id, timestamp, pub_key)

    def validate(self, client_id, timestamp, expected_key):
        """Return true if the header parts are valid.

        If not then AuthenticationError is raised.
        Check if timestamp exceeds a default 5 minute window.
        """
        current_timestamp = time.time()
        if not 0 <= (current_timestamp - int(timestamp)) <= self.timeout:
            return False, TimeoutError('Invalid timestamp.')

        pub_key = utils.generate_public_key(client_id, timestamp, self.secret_key)
        if pub_key != expected_key:
            return False, AuthenticationError('Invalid public key.')
        return True, client_id

    def get_params_value(self, request, name, default=None):
        params = self.get_params(request) or {}
        return params.get(name, default)

    def get_data_value(self, request, name, default=None):
        data = self.get_data(request) or {}
        return data.get(name, default)

    def get_headers(self, request):
        raise NotImplementedError()

    def get_params(self, request):
        raise NotImplementedError()

    def get_data(self, request):
        raise NotImplementedError()

    def get_access_key(self, request):
        return self.get_params_value(
            request,
            self.access_key_name,
            self.get_data_value(request, self.access_key_name)
        )

    def _authenticate(self, request):
        access_key = self.get_access_key(request)
        if access_key is None:
            headers = self.get_headers(request)
            authed_info = self.validate_headers(headers)
        else:
            authed_info = self.validate_access_key(access_key)
        return authed_info

    def authenticate(self, request, raise_exception=True, *args, **kwargs):
        authed_info = self._authenticate(request)
        if authed_info is not None:
            success, id_or_exc = authed_info
            if not success and raise_exception:
                raise id_or_exc
            return success, id_or_exc



from __future__ import absolute_import, print_function, unicode_literals

import os
import time
import unittest
from ssr import BaseAuthentication, random, utils
from ssr.exceptions import AuthenticationError

try:
    import mock
except ImportError:
    from unittest import mock


class ExampleAuthentication(BaseAuthentication):

    def __init__(self):
        self.secret_key = os.environ.get('APP_SECRET_KEY')

    def get_headers(self, request):
        return request.headers

    def get_params(self, request):
        return request.params

    def get_data(self, request):
        return request.data


class AuthenticationTests(unittest.TestCase):

    def setUp(self):
        self.timestamp = int(time.time())
        self.client_id = random.client_id()
        self.secret_key = os.environ.get('APP_SECRET_KEY', random.secret())

        os.environ.setdefault('APP_SECRET_KEY', self.secret_key)
        self.auth = ExampleAuthentication()
        self.request = mock.MagicMock()
        self.request.params = {}
        self.request.data = {}
        self.request.headers = {}

    def prep_headers(self, **kwargs):
        parts = self.valid_parts()
        parts['authorization'] = '{} {}'.format(
            self.auth.auth_type,
            parts['pub_key']
        )
        parts.update(kwargs)
        self.request.headers.update(parts)

    def valid_parts(self):
        return {
            'client_id': self.client_id,
            'timestamp': self.timestamp,
            'pub_key': utils.generate_public_key(
                self.client_id,
                self.timestamp,
                self.secret_key
            )
        }

    def valid_access_token(self):
        return '{client_id}:{timestamp}:{pub_key}'.format(**self.valid_parts())

    def test_authenticate_invalid_timestamps(self):
        now = int(time.time())
        self.prep_headers(timestamp=now + 1000)
        self.assertRaises(
            AuthenticationError,
            self.auth.authenticate,
            self.request
        )
        self.prep_headers(timestamp=now - 1000)
        self.assertRaises(
            AuthenticationError,
            self.auth.authenticate,
            self.request
        )

    def test_authenticate_invalid_authorization_type(self):
        self.prep_headers(authorization='bad')
        self.assertIsNone(self.auth.authenticate(self.request))

    def test_successful_authentication(self):
        self.prep_headers()
        self.assertTrue(self.auth.authenticate(self.request)[0])

    def test_invalid_pub_key(self):
        self.prep_headers(authorization='SSR badtoken')
        self.assertRaises(
            AuthenticationError,
            self.auth.authenticate,
            self.request
        )

    def test_invalid_pub_key_too_many_parts(self):
        self.prep_headers(authorization='SSR bad token')
        self.assertRaises(
            AuthenticationError,
            self.auth.authenticate,
            self.request
        )

    def test_valid_access_token(self):
        access_key = '{client_id}:{timestamp}:{pub_key}'.format(
            **self.valid_parts()
        )
        self.request.params['key'] = access_key
        self.assertTrue(self.auth.authenticate(self.request)[0])

    def test_invalid_access_token(self):
        access_key = 'foo'
        self.request.params['key'] = access_key
        self.assertRaises(
            AuthenticationError,
            self.auth.authenticate,
            self.request
        )

    def test_base_authentication_get_headers(self):
        self.assertRaises(
            NotImplementedError,
            BaseAuthentication().get_headers,
            self.request
        )

    def test_base_authentication_get_params(self):
        self.assertRaises(
            NotImplementedError,
            BaseAuthentication().get_params,
            self.request
        )

    def test_base_authentication_get_data(self):
        self.assertRaises(
            NotImplementedError,
            BaseAuthentication().get_data,
            self.request
        )

    def test_authenticate_header(self):
        self.assertIsNone(
            BaseAuthentication().authenticate_header(self.request)
        )



from __future__ import absolute_import, print_function, unicode_literals

import time
import unittest
from ssr import random
from ssr import Client


class ClientTests(unittest.TestCase):

    def test_client_headers(self):
        client_id = '12345'
        secret_key = '67890'
        client = Client(client_id, secret_key)
        self.assertEqual(client.headers['client_id'], client_id)
        self.assertTrue(client.headers['authorization'].startswith(client.auth_type))

    def test_client_generate_pub_key(self):
        client = Client('id', random.secret())
        timestamp = int(time.time())
        key1 = client.generate_pub_key(timestamp)
        self.assertIsInstance(key1, str)
        key2 = client.generate_pub_key(timestamp)
        self.assertEqual(key1, key2)

    def test_repr(self):
        client = Client('bar', 'foo')
        self.assertEqual(repr(client), '<Client: bar>')



from __future__ import absolute_import, print_function, unicode_literals

import responses

from unittest import TestCase
from ssr import random
from ssr.sessions import Session

try:
    from unittest import mock
except ImportError:
    import mock


class SessionTests(TestCase):

    def setUp(self):
        self.timeout = 1000
        self.session = Session(
            secret_key=random.secret(),
            auth_type='Test',
            auth_timeout=self.timeout
        )

    def test_access_key(self):
        self.assertEqual(
            self.session.access_key,
            self.session.ssr_client.access_key
        )

    def test_timeout(self):
        self.assertEqual(self.session.ssr_client.timeout, self.timeout)

    @responses.activate
    def test_request(self):
        with mock.patch('ssr.sessions.Client.generate_headers') as generate_headers:
            generate_headers.return_value = {'foo': 'bar'}
            responses.add(responses.GET, 'https://httpbin.org/anything')
            self.session.get('https://httpbin.org/anything')
            self.assertTrue(generate_headers.called)



from __future__ import absolute_import, print_function, unicode_literals

from .auth import BaseAuthentication  # noqa: F401
from .clients import Client  # noqa: F401
from .sessions import Session  # noqa: F401



from __future__ import absolute_import, print_function, unicode_literals

import hashlib


def generate_public_key(client_id, timestamp, secret_key):
    """Generate a SHA256 encoded hash using the client id, timestamp and secret key.

    Return the string public key.
    """
    key = '{}{}{}'.format(client_id, secret_key, timestamp)
    return hashlib.sha256(key.encode('utf-8')).hexdigest()


class SecretKeyInterface(object):

    _secret_key = None

    @property
    def secret_key(self):
        if self._secret_key is None:
            self._secret_key = self.get_secret_key()
        return self._secret_key

    @secret_key.setter
    def secret_key(self, value):
        self._secret_key = value

    def get_secret_key(self):
        """Override or set secret_key."""
        pass
